import json

from ..nodes import AbstractNode

class TraceItem:
    """Represents a trace item.
    https://github.com/ProjetPP/Documentation/blob/master/module-communication.md#format-of-a-trace-item
    """
    __slots__ = ('module', 'tree', 'measures')

    def __init__(self, module, tree, measures):
        self.module = module
        self.tree = tree
        self.measures = measures

    @classmethod
    def from_json(cls, data):
        assert isinstance(data, str)
        data = json.loads(data)
        return cls.from_dict(data)

    @classmethod
    def from_dict(cls, data):
        assert isinstance(data, dict)
        return cls(data['module'],
                   AbstractNode.from_dict(data['tree']),
                   data['measures'])

    def as_dict(self):
        return {'module': self.module,
                'tree': self.tree.as_dict(),
                'measures': self.measures,
               }

    def __repr__(self):
        return '<PPP traceitem module=%r, tree=%r, measures=%r>' % \
                (self.module, self.tree, self.measures)

    def __eq__(self, other):
        return self.module == other.module and \
                self.tree == other.tree and \
                self.measures == other.measures

    def __hash__(self):
        return hash((self.module, self.tree, self.measures))



"""Classes representing the nodes of data model of the Projet Pensées Profondes."""

from .abstractnode import AbstractNode
from .triple import Triple
from .missing import Missing
from .resource import *
from .sentence import Sentence
from .list import List
from .list_operators import Union, Intersection, And, Or, First, Last, Sort



"""Contains operators that work on lists."""

from .abstractnode import register, AbstractNode
from .list import List

def to_abstract_node(x):
    assert isinstance(x, (AbstractNode, dict)), x
    return x if isinstance(x, AbstractNode) else AbstractNode.from_dict(x)

class ListNodeOperator(AbstractNode):
    """Base class for list operators.
    https://github.com/ProjetPP/Documentation/blob/master/data-model.md#union-intersection-and-or-first-and-last
    """
    __slots__ = ()
    _possible_attributes = ('list',)

    def _check_attributes(self, attributes):
        super(ListNodeOperator, self)._check_attributes(attributes)
        if not isinstance(attributes['list'], AbstractNode):
            raise TypeError('The “list” argument of the %s constructor '
                            'should be a node, not %r' %
                            (self.__class__.__name__, attributes['list']))

class ListOperator(AbstractNode):
    """Base class for list operators.
    https://github.com/ProjetPP/Documentation/blob/master/data-model.md#union-intersection-and-or-first-and-last
    """
    __slots__ = ()
    _possible_attributes = ('list',)

    def _check_attributes(self, attributes):
        super(ListOperator, self)._check_attributes(attributes)
        if not isinstance(attributes['list'], list):
            raise TypeError('The “list” argument of the %s constructor '
                            'should be an iterable, not %r' %
                            (self.__class__.__name__, attributes['list']))

    def _parse_attributes(self, attributes):
        L = attributes['list']
        assert hasattr(L, '__iter__') and not isinstance(L, AbstractNode)
        if all(isinstance(x, dict) for x in L):
            L = [{'type': 'list', 'list': [x]} if x['type'] != 'list' else x
                 for x in L]
        else:
            assert not any(isinstance(x, dict) for x in L)
        L = tuple(to_abstract_node(l) for l in L)
        attributes['list'] = L
        super(ListOperator, self)._parse_attributes(attributes)

    def as_dict(self):
        d = super().as_dict()
        d['list'] = [x.as_dict() for x in self.list]
        return d

class PredicateListOperator(ListNodeOperator):
    __slots__ = ()
    _possible_attributes = ('list', 'predicate')



@register
class Union(ListOperator):
    __slots__ = ()
    _type = 'union'
@register
class Intersection(ListOperator):
    __slots__ = ()
    _type = 'intersection'
@register
class And(ListOperator):
    __slots__ = ()
    _type = 'and'
@register
class Or(ListOperator):
    __slots__ = ()
    _type = 'or'
@register
class First(ListNodeOperator):
    __slots__ = ()
    _type = 'first'
@register
class Last(ListNodeOperator):
    __slots__ = ()
    _type = 'last'

@register
class Sort(PredicateListOperator):
    __slots__ = ()
    _type = 'sort'



"""Contains a parser for the “human-writable” notation of the datamodel."""

from ply import lex, yacc
import collections

from .nodes import List, Triple, Missing, Resource, Or, And, Intersection, Union

__all__ = ['ParseError', 'parse_triples']

class ParseError(Exception):
    pass

tokens = tuple(('RESOURCE MISSING L_PAREN R_PAREN L_BRACK R_BRACK COMMA '
                'AND OR UNION INTERSECTION').split(' '))
t_L_PAREN = r'\('
t_R_PAREN = r'\)'
t_L_BRACK = r'\['
t_R_BRACK = r'\]'
t_COMMA = r','
t_AND = r'/\\'
t_OR = r'\\/'
t_UNION = r'∪'
t_INTERSECTION = r'∩'
def t_MISSING(t):
    r'\?'
    t.value = Missing()
    return t
def t_RESOURCE(t):
    r'([^()\[\]?,\\/∪∩]+|"([^"\\]|\\.)*")'
    # TODO: remove / from the regexp
    t.value = Resource(t.value.strip())
    return t
#forbidden = '|'.join((
#    r'\/', r'/\\', '\(', '\)', '\[', '\]', r'(?<!\\)"'))
#t_RESOURCE.__doc__ = '(?!(%s)*)' % (forbidden,)

t_ignore = ' '

def t_error(t):
    raise ParseError('Illegal string `%s`' % t.value)

lex.lex()


def p_triple(t):
    """triple : L_PAREN expression COMMA expression COMMA expression R_PAREN"""
    t[0] = Triple(t[2], t[4], t[6])

def p_list_body_singleton(t):
    """list_body : expression"""
    t[0] = [t[1]]

def p_list_body_comma(t):
    """list_body : list_body COMMA expression"""
    t[0] = t[1]
    t[0].append(t[3])

def p_list(t):
    """list : L_BRACK list_body R_BRACK"""
    t[0] = List(t[2])

def p_list_empty(t):
    """list : L_BRACK R_BRACK"""
    t[0] = List([])

def p_conjonction_base(t):
    """conjonction_base : RESOURCE
                        | MISSING
                        | triple
                        | list"""
    t[0] = t[1]
def p_conjonction_expr(t):
    """conjonction : L_PAREN expression R_PAREN"""
    t[0] = [t[2]]
def p_conjonction_empty(t):
    """conjonction : conjonction_base"""
    t[0] = [t[1]]
def p_conjonction(t):
    """conjonction : conjonction AND conjonction_base"""
    t[0] = t[1]
    t[0].append(t[3])

def p_disjonction_empty(t):
    """disjonction : conjonction"""
    assert isinstance(t[1], list), t[1]
    t[0] = [And(t[1])]
def p_disjonction(t):
    """disjonction : disjonction OR conjonction"""
    t[0] = t[1]
    t[0].append(And(t[3]))


def p_intersection_empty(t):
    """intersection : disjonction"""
    assert isinstance(t[1], list), t[1]
    t[0] = [Or(t[1])]
def p_intersection(t):
    """intersection : intersection INTERSECTION disjonction"""
    t[0] = t[1]
    t[0].append(Or(t[3]))

def p_union_empty(t):
    """union : intersection"""
    assert isinstance(t[1], list), t[1]
    t[0] = [Intersection(t[1])]
def p_union(t):
    """union : union UNION intersection"""
    t[0] = t[1]
    t[0].append(Intersection(t[3]))

def p_expression(t):
    """expression : union"""
    assert isinstance(t[1], list)
    t[0] = Union(t[1])


def p_error(t):
    raise ParseError("Syntax error at '%s'" % t.value)

parser = yacc.yacc(start='expression')

def normalize(tree):
    if isinstance(tree, (List, Or, And, Union, Intersection)) and \
            len(tree.list) == 1:
        return normalize(tree.list[0])
    elif isinstance(tree, (List, Or, And, Union, Intersection)):
        return tree.__class__(list(map(normalize, tree.list)))
    elif isinstance(tree, Triple):
        return Triple(normalize(tree.subject),
                      normalize(tree.predicate),
                      normalize(tree.object))
    else:
        return tree

def parse_triples(sentence):
    return normalize(parser.parse(sentence))



"""Module containing exception classes."""

class PPPDatamodelException(Exception):
    """Base class for PPP exceptions."""
    pass
class AttributeNotProvided(PPPDatamodelException):
    """Exception raised when a mandatory attribute is not given to
    a node constructor."""
    pass
class UnknownNodeType(PPPDatamodelException):
    """Exception raised when a node type has no constructor/class."""
    pass



import json

from .traceitem import TraceItem
from ..nodes import AbstractNode

class Request:
    """Represents a request.
    https://github.com/ProjetPP/Documentation/blob/master/module-communication.md#request
    """
    __slots__ = ('id', 'language', 'tree', 'measures', 'trace')

    def __init__(self, id, language, tree, measures={}, trace=[]):
        self.id = id
        assert isinstance(tree, str) or isinstance(tree, AbstractNode)
        self.tree = tree
        self.language = language
        self.measures = measures
        self.trace = [x if isinstance(x, TraceItem) else TraceItem.from_dict(x)
                      for x in trace]

    def __repr__(self):
        return '<ppp_datamodel.Request(%s)>' % \
                ', '.join(map(lambda x:'%s=%r' % (x, getattr(self, x)),
                              self.__slots__))

    def __eq__(self, other):
        if not isinstance(other, Request):
            return False
        return self.id == other.id and \
                self.language == other.language and \
                self.tree == other.tree

    def __hash__(self):
        return hash((self.id, self.language, self.tree, self.trace))

    @classmethod
    def from_json(cls, data):
        assert isinstance(data, str)
        data = json.loads(data)
        return cls.from_dict(data)

    @classmethod
    def from_dict(cls, data):
        assert isinstance(data, dict)
        tree = data['tree']
        if isinstance(tree, dict):
            tree = AbstractNode.from_dict(tree)
        return cls(data['id'],
                   data['language'],
                   tree,
                   data['measures'],
                   data['trace'])


    def as_dict(self):
        tree = self.tree
        if isinstance(tree, AbstractNode):
            tree = tree.as_dict()
        d = {'id': self.id, 'language': self.language,
             'measures': self.measures,
             'trace': [x.as_dict() for x in self.trace],
             'tree': tree}
        return d
    def as_json(self):
        return json.dumps(self.as_dict())



import json

from ..nodes import AbstractNode
from .traceitem import TraceItem

class Response:
    """Represents a response.
    https://github.com/ProjetPP/Documentation/blob/master/module-communication.md#response
    """
    __slots__ = ('language', 'tree', 'measures', 'trace')

    def __init__(self, language, tree, measures, trace):
        if isinstance(tree, dict):
            tree = AbstractNode.from_dict(tree)
        elif isinstance(tree, str):
            tree = AbstractNode.from_json(tree)
        self.language = language
        self.measures = measures
        self.tree = tree
        self.trace = [x if isinstance(x, TraceItem) else TraceItem.from_dict(x)
                      for x in trace]

    def __repr__(self):
        return '<PPP response language=%r, tree=%r, measures=%r, trace=%r>'%\
                (self.language, self.tree, self.measures, self.trace)

    def __eq__(self, other):
        if not isinstance(other, Response):
            return False
        return self.language == other.language and \
                self.tree == other.tree and \
                self.measures == other.measures and \
                self.trace == other.trace

    def __hash__(self):
        return hash((self.language, self.tree, self.measures, self.trace))

    @classmethod
    def from_json(cls, data):
        data = json.loads(data)
        return cls.from_dict(data)

    @classmethod
    def from_dict(cls, data):
        return cls(data['language'], data['tree'], data['measures'],
                   list(map(TraceItem.from_dict, data['trace'])))

    def as_dict(self):
        return {'language': self.language,
                'tree': self.tree.as_dict(),
                'measures': self.measures,
                'trace': [x.as_dict() for x in self.trace]
               }
    def as_json(self):
        return json.dumps(self.as_dict())
